@@ -65,6 +65,7 @@ enum {
 	MOVE_COPY_ITEMS,
 	HANDLE_URL,
 	HANDLE_URI_LIST,
+	HANDLE_TEXT,
 	LAST_SIGNAL
 };
 
@@ -80,7 +81,7 @@ GNOME_CLASS_BOILERPLATE (NautilusTreeViewDragDest,
 static const GtkTargetEntry drag_types [] = {
 	{ NAUTILUS_ICON_DND_GNOME_ICON_LIST_TYPE, 0, NAUTILUS_ICON_DND_GNOME_ICON_LIST },
 	{ NAUTILUS_ICON_DND_URI_LIST_TYPE, 0, NAUTILUS_ICON_DND_URI_LIST },
-	{ NAUTILUS_ICON_DND_URL_TYPE, 0, NAUTILUS_ICON_DND_URL }
+	{ NAUTILUS_ICON_DND_URL_TYPE, 0, NAUTILUS_ICON_DND_URL },
 	/* FIXME: Should handle emblems once the list view supports them */
 };
 
@@ -384,6 +385,17 @@ get_drop_action (NautilusTreeViewDragDest *dest,
 		
 	case NAUTILUS_ICON_DND_URI_LIST :
 		return context->suggested_action;
+
+	case NAUTILUS_ICON_DND_TEXT:
+		drop_target = get_drop_target (dest, path);
+
+		if (!drop_target) {
+			return 0;
+		}
+
+		g_free (drop_target);
+
+		return GDK_ACTION_COPY;
 	}
 
 	return 0;
@@ -579,6 +591,26 @@ receive_dropped_uri_list (NautilusTreeViewDragDest *dest,
 		       x, y);
 }
 
+static void
+receive_dropped_text (NautilusTreeViewDragDest *dest,
+		      GdkDragContext *context,
+		      int x, int y)
+{
+	char *text;
+
+	if (!dest->details->drag_data) {
+		return;
+	}
+
+	text = gtk_selection_data_get_text (dest->details->drag_data);
+	g_signal_emit (dest, signals[HANDLE_TEXT], 0,
+		       (char *) text,
+		       context->action,
+		       x, y);
+	g_free (text);
+}
+
+
 static void
 receive_dropped_url (NautilusTreeViewDragDest *dest,
 		     GdkDragContext *context,
@@ -635,6 +667,10 @@ drag_data_received_callback (GtkWidget *widget,
 			receive_dropped_uri_list (dest, context, x, y);
 			success = TRUE;
 			break;
+		case NAUTILUS_ICON_DND_TEXT:
+			receive_dropped_text (dest, context, x, y);
+			success = TRUE;
+			break;
 		}
 
 		dest->details->drop_occurred = FALSE;
@@ -762,11 +798,11 @@ nautilus_tree_view_drag_dest_class_init (NautilusTreeViewDragDestClass *class)
 					       move_copy_items),
 			      NULL, NULL,
 			      
-			      nautilus_marshal_VOID__POINTER_STRING_UINT_INT_INT,
+			      nautilus_marshal_VOID__POINTER_STRING_ENUM_INT_INT,
 			      G_TYPE_NONE, 5,
 			      G_TYPE_POINTER,
 			      G_TYPE_STRING,
-			      G_TYPE_UINT,
+			      GDK_TYPE_DRAG_ACTION,
 			      G_TYPE_INT,
 			      G_TYPE_INT);
 	signals[HANDLE_URL] =
@@ -776,10 +812,10 @@ nautilus_tree_view_drag_dest_class_init (NautilusTreeViewDragDestClass *class)
 			      G_STRUCT_OFFSET (NautilusTreeViewDragDestClass, 
 					       handle_url),
 			      NULL, NULL,
-			      nautilus_marshal_VOID__STRING_INT_INT_INT,
+			      nautilus_marshal_VOID__STRING_ENUM_INT_INT,
 			      G_TYPE_NONE, 4,
 			      G_TYPE_STRING,
-			      G_TYPE_INT,
+			      GDK_TYPE_DRAG_ACTION,
 			      G_TYPE_INT,
 			      G_TYPE_INT);
 	signals[HANDLE_URI_LIST] =
@@ -789,10 +825,23 @@ nautilus_tree_view_drag_dest_class_init (NautilusTreeViewDragDestClass *class)
 			      G_STRUCT_OFFSET (NautilusTreeViewDragDestClass, 
 					       handle_uri_list),
 			      NULL, NULL,
-			      nautilus_marshal_VOID__STRING_INT_INT_INT,
+			      nautilus_marshal_VOID__STRING_ENUM_INT_INT,
 			      G_TYPE_NONE, 4,
 			      G_TYPE_STRING,
+			      GDK_TYPE_DRAG_ACTION,
 			      G_TYPE_INT,
+			      G_TYPE_INT);
+	signals[HANDLE_TEXT] =
+		g_signal_new ("handle_text",
+			      G_TYPE_FROM_CLASS (class),
+			      G_SIGNAL_RUN_LAST,
+			      G_STRUCT_OFFSET (NautilusTreeViewDragDestClass, 
+					       handle_text),
+			      NULL, NULL,
+			      nautilus_marshal_VOID__STRING_ENUM_INT_INT,
+			      G_TYPE_NONE, 4,
+			      G_TYPE_STRING,
+			      GDK_TYPE_DRAG_ACTION,
 			      G_TYPE_INT,
 			      G_TYPE_INT);
 }
@@ -803,6 +852,7 @@ NautilusTreeViewDragDest *
 nautilus_tree_view_drag_dest_new (GtkTreeView *tree_view)
 {
 	NautilusTreeViewDragDest *dest;
+	GtkTargetList *targets;
 	
 	dest = g_object_new (NAUTILUS_TYPE_TREE_VIEW_DRAG_DEST, NULL);
 
@@ -812,8 +862,11 @@ nautilus_tree_view_drag_dest_new (GtkTreeView *tree_view)
 	
 	gtk_drag_dest_set (GTK_WIDGET (tree_view),
 			   0, drag_types, G_N_ELEMENTS (drag_types),
-			   GDK_ACTION_MOVE | GDK_ACTION_COPY | GDK_ACTION_LINK | GDK_ACTION_ASK);	
-	
+			   GDK_ACTION_MOVE | GDK_ACTION_COPY | GDK_ACTION_LINK | GDK_ACTION_ASK);
+
+	targets = gtk_drag_dest_get_target_list (GTK_WIDGET (tree_view));
+	gtk_target_list_add_text_targets (targets, NAUTILUS_ICON_DND_TEXT);
+
 	g_signal_connect_object (tree_view,
 				 "drag_motion",
 				 G_CALLBACK (drag_motion_callback),
