@@ -21,9 +21,6 @@
  *
  */
 
-/* This object exports the bonobo context menus and property pages
- * using the new extension interface. */ 
-
 #include <config.h>
 #include "fm-ditem-page.h"
 
@@ -66,7 +63,7 @@ enum {
 	TARGET_URI_LIST
 };
 
-static GtkTargetEntry target_table[] = {
+static const GtkTargetEntry target_table[] = {
         { "text/uri-list",  0, TARGET_URI_LIST }
 };
 
