@@ -36,6 +36,7 @@
 #include <eel/eel-vfs-extensions.h>
 #include <eel/eel-glib-extensions.h>
 #include <eel/eel-gtk-macros.h>
+#include <gdk/gdkcursor.h>
 #include <gdk/gdkkeysyms.h>
 #include <gtk/gtkdialog.h>
 #include <gtk/gtkbindings.h>
@@ -100,12 +101,16 @@ struct FMListViewDetails {
 
 	GtkTreePath *new_selection_path;   /* Path of the new selection after removing a file */
 
+	GtkTreePath *hover_path;
+
 	guint drag_button;
 	int drag_x;
 	int drag_y;
 
 	gboolean drag_started;
 
+	gboolean ignore_button_release;
+
 	gboolean row_selected_on_button_down;
 
 	/* typeahead selection state */
@@ -408,8 +413,11 @@ motion_notify_callback (GtkWidget *widget,
 {
 	FMListView *view;
 	GdkDragContext *context;
+	GdkCursor *cursor;
 	GList *ref_list;
 	GdkPixbuf *pixbuf;	
+	GtkTreePath *last_hover_path;
+	GtkTreeIter iter;
 	
 	view = FM_LIST_VIEW (callback_data);
 	
@@ -417,6 +425,45 @@ motion_notify_callback (GtkWidget *widget,
 		return FALSE;
 	}
 
+	if (click_policy_auto_value == NAUTILUS_CLICK_POLICY_SINGLE) {
+		last_hover_path = view->details->hover_path;
+
+		gtk_tree_view_get_path_at_pos (GTK_TREE_VIEW (widget),
+					       event->x, event->y,
+					       &view->details->hover_path,
+					       NULL, NULL, NULL);
+
+		if (view->details->hover_path != NULL) {
+			cursor = gdk_cursor_new (GDK_HAND2);
+		} else {
+			cursor = NULL;
+		}
+
+		gdk_window_set_cursor (widget->window, cursor);
+
+		/* only redraw if the hover row has changed */
+		if (!(last_hover_path == NULL && view->details->hover_path == NULL) &&
+		    (!(last_hover_path != NULL && view->details->hover_path != NULL) ||
+		     gtk_tree_path_compare (last_hover_path, view->details->hover_path))) {
+			if (last_hover_path) {
+				if (gtk_tree_model_get_iter (GTK_TREE_MODEL (view->details->model),
+							     &iter, last_hover_path)) {
+					gtk_tree_model_row_changed (GTK_TREE_MODEL (view->details->model),
+								    last_hover_path, &iter);
+				}
+			}
+
+			if (view->details->hover_path) {
+				gtk_tree_model_get_iter (GTK_TREE_MODEL (view->details->model),
+							 &iter, view->details->hover_path);
+				gtk_tree_model_row_changed (GTK_TREE_MODEL (view->details->model),
+							    view->details->hover_path, &iter);
+			}
+		}
+
+		gtk_tree_path_free (last_hover_path);
+	}
+
 	if (view->details->drag_button != 0) {
 		if (gtk_drag_check_threshold (widget,
 					      view->details->drag_x,
@@ -455,6 +502,33 @@ motion_notify_callback (GtkWidget *widget,
 	return FALSE;
 }
 
+static gboolean
+leave_notify_callback (GtkWidget *widget,
+		       GdkEventCrossing *event,
+		       gpointer callback_data)
+{
+	FMListView *view;
+	GtkTreeIter iter;
+
+	view = FM_LIST_VIEW (callback_data);
+
+	if (click_policy_auto_value == NAUTILUS_CLICK_POLICY_SINGLE &&
+	    view->details->hover_path != NULL) {
+		if (gtk_tree_model_get_iter (GTK_TREE_MODEL (view->details->model),
+					     &iter, view->details->hover_path)) {
+			gtk_tree_model_row_changed (GTK_TREE_MODEL (view->details->model),
+						    view->details->hover_path, &iter);
+		}
+
+		gtk_tree_path_free (view->details->hover_path);
+		view->details->hover_path = NULL;
+
+		return TRUE;
+	}
+
+	return FALSE;
+}
+
 static void
 do_popup_menu (GtkWidget *widget, FMListView *view, GdkEventButton *event)
 {
@@ -480,6 +554,7 @@ button_press_callback (GtkWidget *widget, GdkEventButton *event, gpointer callba
 	static int click_count = 0;
 	int double_click_time;
 	int expander_size, horizontal_separator;
+	gboolean on_expander;
 
 	view = FM_LIST_VIEW (callback_data);
 	tree_view = GTK_TREE_VIEW (widget);
@@ -515,7 +590,8 @@ button_press_callback (GtkWidget *widget, GdkEventButton *event, gpointer callba
 		return TRUE;
 	}
 
-	
+	view->details->ignore_button_release = FALSE;
+
 	call_parent = TRUE;
 	allow_drag = FALSE;
 	if (gtk_tree_view_get_path_at_pos (tree_view, event->x, event->y,
@@ -568,17 +644,23 @@ button_press_callback (GtkWidget *widget, GdkEventButton *event, gpointer callba
 			if ((event->button == 1 || event->button == 2) &&
 			    ((event->state & GDK_CONTROL_MASK) != 0 ||
 			     (event->state & GDK_SHIFT_MASK) == 0)) {
+				gtk_widget_style_get (widget,
+						      "expander-size", &expander_size,
+						      "horizontal-separator", &horizontal_separator,
+						      NULL);
+				on_expander = (event->x <= horizontal_separator / 2 +
+					       gtk_tree_path_get_depth (path) * expander_size);
+
+				
 				view->details->row_selected_on_button_down = gtk_tree_selection_path_is_selected (selection, path);
 				if (view->details->row_selected_on_button_down) {
-					gtk_widget_style_get (widget,
-						"expander-size", &expander_size,
-						"horizontal-separator", &horizontal_separator,
-						NULL);
-					call_parent = (event->x <= horizontal_separator / 2 +
-						gtk_tree_path_get_depth (path) * expander_size);
+					call_parent = on_expander;
+					view->details->ignore_button_release = call_parent;
 				} else if  ((event->state & GDK_CONTROL_MASK) != 0) {
 					call_parent = FALSE;
 					gtk_tree_selection_select_path (selection, path);
+				} else {
+					view->details->ignore_button_release = on_expander;
 				}
 			}
 		
@@ -636,7 +718,8 @@ button_release_callback (GtkWidget *widget,
 
 	if (event->button == view->details->drag_button) {
 		stop_drag_check (view);
-		if (!view->details->drag_started) {
+		if (!view->details->drag_started &&
+		    !view->details->ignore_button_release) {
 			fm_list_view_did_not_drag (view, event);
 		}
 	}
@@ -1071,6 +1154,13 @@ list_view_handle_uri_list (NautilusTreeViewDragDest *dest, const char *item_uris
 						item_uris, action, x, y);
 }
 
+static void
+list_view_handle_text (NautilusTreeViewDragDest *dest, const char *text,
+		       GdkDragAction action, int x, int y, FMListView *view)
+{
+	fm_directory_view_handle_text_drop (FM_DIRECTORY_VIEW (view),
+					    text, action, x, y);
+}
 
 static void
 move_copy_items_callback (NautilusTreeViewDragDest *dest,
@@ -1144,6 +1234,42 @@ apply_column_order (FMListView *list_view, GList *column_order)
 	nautilus_column_list_free (columns);
 }
 
+static void
+filename_cell_data_func (GtkTreeViewColumn *column,
+			 GtkCellRenderer   *renderer,
+			 GtkTreeModel      *model,
+			 GtkTreeIter       *iter,
+			 FMListView        *view)
+{
+	char *text;
+	GtkTreePath *path;
+	PangoUnderline underline;
+
+	gtk_tree_model_get (model, iter,
+			    view->details->file_name_column_num, &text,
+			    -1);
+
+	if (click_policy_auto_value == NAUTILUS_CLICK_POLICY_SINGLE) {
+		path = gtk_tree_model_get_path (model, iter);
+
+		if (view->details->hover_path == NULL ||
+		    gtk_tree_path_compare (path, view->details->hover_path)) {
+			underline = PANGO_UNDERLINE_NONE;
+		} else {
+			underline = PANGO_UNDERLINE_SINGLE;
+		}
+
+		gtk_tree_path_free (path);
+	} else {
+		underline = PANGO_UNDERLINE_NONE;
+	}
+
+	g_object_set (G_OBJECT (renderer),
+		      "text", text,
+		      "underline", underline,
+		      NULL);
+}
+
 static void
 create_and_set_up_tree_view (FMListView *view)
 {
@@ -1187,6 +1313,8 @@ create_and_set_up_tree_view (FMListView *view)
 				 G_CALLBACK (list_view_handle_url), view, 0);
 	g_signal_connect_object (view->details->drag_dest, "handle_uri_list",
 				 G_CALLBACK (list_view_handle_uri_list), view, 0);
+	g_signal_connect_object (view->details->drag_dest, "handle_text",
+				 G_CALLBACK (list_view_handle_text), view, 0);
 
 	g_signal_connect_object (gtk_tree_view_get_selection (view->details->tree_view),
 				 "changed",
@@ -1196,6 +1324,8 @@ create_and_set_up_tree_view (FMListView *view)
 				 G_CALLBACK (drag_data_get_callback), view, 0);
 	g_signal_connect_object (view->details->tree_view, "motion_notify_event",
 				 G_CALLBACK (motion_notify_callback), view, 0);
+	g_signal_connect_object (view->details->tree_view, "leave_notify_event",
+				 G_CALLBACK (leave_notify_callback), view, 0);
 	g_signal_connect_object (view->details->tree_view, "button_press_event",
 				 G_CALLBACK (button_press_callback), view, 0);
 	g_signal_connect_object (view->details->tree_view, "button_release_event",
@@ -1278,9 +1408,9 @@ create_and_set_up_tree_view (FMListView *view)
 			g_signal_connect (cell, "editing-canceled", G_CALLBACK (cell_renderer_editing_canceled), view);
 			
 			gtk_tree_view_column_pack_start (view->details->file_name_column, cell, TRUE);
-			gtk_tree_view_column_set_attributes (view->details->file_name_column, cell,
-							     "text", 
-							     column_num, NULL);
+			gtk_tree_view_column_set_cell_data_func (view->details->file_name_column, cell,
+								 (GtkTreeCellDataFunc) filename_cell_data_func,
+								 view, NULL);
 			gtk_tree_view_append_column (view->details->tree_view, view->details->file_name_column);
 		} else {		
 			cell = gtk_cell_renderer_text_new ();
@@ -2152,7 +2282,7 @@ fm_list_view_start_renaming_file (FMDirectoryView *view,
 	gtk_tree_view_scroll_to_cell (list_view->details->tree_view,
 				      path,
 				      list_view->details->file_name_column,
-				      FALSE, 0.0, 1.0);
+				      FALSE, 0.0, 0.0);
 	gtk_tree_view_set_cursor (list_view->details->tree_view,
 				  path,
 				  list_view->details->file_name_column,
@@ -2184,22 +2314,39 @@ fm_list_view_start_renaming_file (FMDirectoryView *view,
 static void
 fm_list_view_click_policy_changed (FMDirectoryView *directory_view)
 {
+	GdkWindow *win;
+	GdkDisplay *display;
 	FMListView *view;
+	GtkTreeIter iter;
+	GtkTreeView *tree;
 
-	view = FM_LIST_VIEW (directory_view);
+	/* ensure that we unset the hand cursor and refresh underlined rows */
+	if (click_policy_auto_value == NAUTILUS_CLICK_POLICY_DOUBLE) {
+		view = FM_LIST_VIEW (directory_view);
+		
+		if (view->details->hover_path != NULL) {
+			if (gtk_tree_model_get_iter (GTK_TREE_MODEL (view->details->model),
+						     &iter, view->details->hover_path)) {
+				gtk_tree_model_row_changed (GTK_TREE_MODEL (view->details->model),
+							    view->details->hover_path, &iter);
+			}
 
-	if (click_policy_auto_value == NAUTILUS_CLICK_POLICY_SINGLE) {
-		g_object_set (G_OBJECT (view->details->file_name_cell),
-			      "underline", PANGO_UNDERLINE_SINGLE,
-			      NULL);
-	} else {
-		g_object_set (G_OBJECT (view->details->file_name_cell),
-			      "underline", PANGO_UNDERLINE_NONE,
-			      NULL);
+			gtk_tree_path_free (view->details->hover_path);
+			view->details->hover_path = NULL;
+		}
+
+		tree = view->details->tree_view;
+		if (GTK_WIDGET_REALIZED (GTK_WIDGET (tree))) {
+			win = GTK_WIDGET (tree)->window;
+			gdk_window_set_cursor (win, NULL);
+			
+			display = gtk_widget_get_display (GTK_WIDGET (view));
+			if (display != NULL) {
+				gdk_display_flush (display);
+			}
+		}
+		
 	}
-
-	gtk_tree_model_foreach (GTK_TREE_MODEL (view->details->model),
-				list_view_changed_foreach, NULL);
 }
 
 static void
