@@ -140,10 +140,10 @@ nautilus_navigation_window_instance_init (NautilusNavigationWindow *window)
 	window->details->content_paned = nautilus_horizontal_splitter_new ();
 	gtk_table_attach (GTK_TABLE (NAUTILUS_WINDOW (window)->details->table),
 			  window->details->content_paned,
-			  /* X direction */       /* Y direction */
-			  0, 1,                   3, 4,
-			  GTK_EXPAND | GTK_FILL,  GTK_EXPAND | GTK_FILL,
-			  0,                      0);
+			  /* X direction */                   /* Y direction */
+			  0, 1,                               3, 4,
+			  GTK_EXPAND | GTK_FILL | GTK_SHRINK, GTK_EXPAND | GTK_FILL | GTK_SHRINK,
+			  0,                                  0);
 	gtk_widget_show (window->details->content_paned);
 
 	nautilus_navigation_window_initialize_actions (window);
@@ -154,10 +154,10 @@ nautilus_navigation_window_instance_init (NautilusNavigationWindow *window)
 	window->details->toolbar = toolbar;
 	gtk_table_attach (GTK_TABLE (NAUTILUS_WINDOW (window)->details->table),
 			  toolbar,
-			  /* X direction */       /* Y direction */
-			  0, 1,                   1, 2,
-			  GTK_EXPAND | GTK_FILL,  0,
-			  0,                      0);
+			  /* X direction */                   /* Y direction */
+			  0, 1,                               1, 2,
+			  GTK_EXPAND | GTK_FILL | GTK_SHRINK, 0,
+			  0,                                  0);
 	gtk_widget_show (toolbar);
 
 	nautilus_navigation_window_initialize_toolbars (window);
@@ -246,10 +246,10 @@ nautilus_navigation_window_instance_init (NautilusNavigationWindow *window)
 
 	gtk_table_attach (GTK_TABLE (NAUTILUS_WINDOW (window)->details->table),
 			  location_bar,
-			  /* X direction */       /* Y direction */
-			  0, 1,                   2, 3,
-			  GTK_EXPAND | GTK_FILL,  0,
-			  0,                      0);
+			  /* X direction */                    /* Y direction */
+			  0, 1,                                2, 3,
+			  GTK_EXPAND | GTK_FILL | GTK_SHRINK,  0,
+			  0,                                   0);
 
 	eel_preferences_add_callback_while_alive (NAUTILUS_PREFERENCES_ALWAYS_USE_LOCATION_ENTRY,
 						  always_use_location_entry_changed,
@@ -270,14 +270,47 @@ always_use_location_entry_changed (gpointer callback_data)
 	}
 }
 
+static int
+bookmark_list_get_uri_index (GList *list,
+			     const char *uri)
+{
+	NautilusBookmark *bookmark;
+	GList *l;
+	char *tmp;
+	int i;
+
+	g_return_val_if_fail (uri != NULL, -1);
+
+	for (i = 0, l = list; l != NULL; i++, l = l->next) {
+		bookmark = NAUTILUS_BOOKMARK (l->data);
+
+		tmp = nautilus_bookmark_get_uri (bookmark);
+		if (strcmp (tmp, uri) == 0) {
+			g_free (tmp);
+			return i;
+		}
+		g_free (tmp);
+	}
+
+	return -1;
+}
 
 static void
 path_bar_location_changed_callback (GtkWidget *widget,
 				    const char *uri,
 				    NautilusNavigationWindow *window)
 {
+	int i;
+
 	g_assert (NAUTILUS_IS_NAVIGATION_WINDOW (window));
-	nautilus_window_go_to (NAUTILUS_WINDOW (window), uri);
+
+	/* check whether we already visited the target location */
+	i = bookmark_list_get_uri_index (window->back_list, uri);
+	if (i >= 0) {
+		nautilus_navigation_window_back_or_forward (window, TRUE, i);
+	} else {
+		nautilus_window_go_to (NAUTILUS_WINDOW (window), uri);
+	}
 }
 
 
@@ -751,6 +784,12 @@ real_set_title (NautilusWindow *window, const char *title)
 	g_free (full_title);
 }
 
+static char *
+real_get_icon_name (NautilusWindow *window)
+{
+	return g_strdup ("file-manager");
+}
+
 static void
 zoom_level_changed_callback (NautilusView *view,
                              NautilusNavigationWindow *window)
@@ -1230,6 +1269,7 @@ nautilus_navigation_window_class_init (NautilusNavigationWindowClass *class)
 	NAUTILUS_WINDOW_CLASS (class)->set_throbber_active = real_set_throbber_active;
 	NAUTILUS_WINDOW_CLASS (class)->prompt_for_location = real_prompt_for_location;
 	NAUTILUS_WINDOW_CLASS (class)->set_title = real_set_title;
+	NAUTILUS_WINDOW_CLASS (class)->get_icon_name = real_get_icon_name;
 	NAUTILUS_WINDOW_CLASS(class)->get_default_size = real_get_default_size;
 	NAUTILUS_WINDOW_CLASS (class)->close = real_window_close;
 }
