@@ -1178,8 +1178,10 @@ add_new_pattern (NautilusPropertyBrowser *property_browser)
 static void
 add_color_to_file (NautilusPropertyBrowser *property_browser, const char *color_spec, const char *color_name)
 {
-	xmlNodePtr cur_node, new_color_node;
+	xmlNodePtr cur_node, new_color_node, children_node;
 	xmlDocPtr document;
+	xmlChar *child_color_name;
+	gboolean color_name_exists = FALSE;
 
 	document = read_browser_xml (property_browser);
 	if (document == NULL) {
@@ -1189,13 +1191,33 @@ add_color_to_file (NautilusPropertyBrowser *property_browser, const char *color_
 	/* find the colors category */
 	cur_node = get_color_category (document);
 	if (cur_node != NULL) {
-		/* add a new color node */
-		new_color_node = xmlNewChild (cur_node, NULL, "color", NULL);
-		xmlNodeSetContent (new_color_node, color_spec);
-		xmlSetProp (new_color_node, "local", "1");
-		xmlSetProp (new_color_node, "name", color_name);
+		/* check if theres already a color whith that name */
+		children_node = cur_node->xmlChildrenNode;
+		while (children_node != NULL) {
+			child_color_name = xmlGetProp (children_node, "name");	
+			if (xmlStrcmp (color_name, child_color_name) == 0) {
+				color_name_exists = TRUE;
+				xmlFree (child_color_name);
+				break;
+			}
+			xmlFree (child_color_name);
+
+			children_node = children_node->next;
+		}
 		
-		write_browser_xml (property_browser, document);
+		/* add a new color node */
+		if (!color_name_exists) {
+			new_color_node = xmlNewChild (cur_node, NULL, "color", NULL);
+			xmlNodeSetContent (new_color_node, color_spec);
+			xmlSetProp (new_color_node, "local", "1");
+			xmlSetProp (new_color_node, "name", color_name);
+				
+			write_browser_xml (property_browser, document);
+		} else {
+			eel_show_error_dialog (_("The color cannot be installed."),
+					       _("Sorry, but you must specify an unused color name for the new color."), 
+			                       _("Couldn't Install Color"), GTK_WINDOW (property_browser));
+		}
 	}
 	
 	xmlFreeDoc (document);
@@ -1347,6 +1369,8 @@ emblem_dialog_clicked (GtkWidget *dialog, int which_button, NautilusPropertyBrow
 						       stripped_keyword,
 						       stripped_keyword,
 						       GTK_WINDOW (property_browser));
+		if (pixbuf != NULL)
+			g_object_unref (pixbuf);
 
 		nautilus_emblem_refresh_list ();
 		
@@ -1728,6 +1752,7 @@ make_properties_from_directories (NautilusPropertyBrowser *property_browser)
 		object_pixbuf = gdk_pixbuf_new_from_file (path, NULL);
 		g_free (path);
 		property_image = labeled_image_new (_("Erase"), object_pixbuf, "erase", PANGO_SCALE_LARGE);
+		g_object_unref (object_pixbuf);
 		eel_labeled_image_set_fixed_image_height (EEL_LABELED_IMAGE (property_image), MAX_EMBLEM_HEIGHT);
 
 		gtk_container_add (GTK_CONTAINER (image_table), property_image);
